from __future__ import print_function, absolute_import

import HTSeq
#from from deprecation import deprecated
from lazy import lazy

from pyreference.genomic_region import GenomicRegion
from pyreference.settings import START, END, IS_CODING, CHROM, STRAND
from pyreference.utils.genomics_utils import GenomicInterval_from_directional, dict_to_iv


class NotOnTranscriptException(Exception):
    """ For when a transcription position is not on a transcript """
    pass


class Transcript(GenomicRegion):
    def __init__(self, *args, **kwargs):
        gene = kwargs.pop("gene", None)
        super(Transcript, self).__init__(*args, **kwargs)
        self.gene = gene
    
    def get_gene_id(self):
        return self.gene.get_id() 

    @property
    def is_coding(self):
        return self._dict[IS_CODING]

    def get_representative_transcript(self):
        return self


    def get_features_length(self, feature_type):
        length = 0
        for feature in self.get_features_in_stranded_order(feature_type):
            length += feature[END] - feature[START] 
        return length

    #@deprecated(details="Use get_features_in_stranded_order")
    def get_features(self, feature_type):
        """ returns list of HTSeq.GenomicFeature """
        genomic_features = []
        for f in self.get_features_in_stranded_order(feature_type):
            iv = dict_to_iv(f)
            genomic_feature = HTSeq.GenomicFeature(self.get_id(), feature_type, iv)
            genomic_features.append(genomic_feature)
        
        return genomic_features 


    def get_features_in_stranded_order(self, feature_type):
        """features returned sorted 5' -> 3' """

        is_reversed = self._dict["strand"] == '-'
        if is_reversed:
            stranded_start = END
        else:
            stranded_start = START
            
        features_by_type = self._dict["features_by_type"]
        features = features_by_type.get(feature_type, [])
        if features:
            # Need to add this as not in there by default
            transcript_chrom = self._dict[CHROM]
            transcript_strand = self._dict[STRAND]

            for f in features:
                f[CHROM] = transcript_chrom
                f[STRAND] = transcript_strand

            features = sorted(features, key=lambda x : x[stranded_start], reverse=is_reversed)
        return features

    @lazy
    def length(self):
        return self.get_features_length("exon")
    
    #@deprecated(details="Use Transcript.length")
    def get_transcript_length(self):
        return self.length

    def get_sequence_from_features(self, feature_type):
        features = self.get_features_in_stranded_order(feature_type)
        return self.reference.get_sequence_from_features(features)

    def get_transcript_sequence(self):
        return self.get_sequence_from_features("exon")

    @lazy
    def exons(self):
        """ Returns list of exon features """
        return self.get_features("exon")
    
    @lazy
    def threeputr(self):
        """ Returns the exon regions which contain 3'UTR as list of features """
        return self.get_features("3PUTR") 
    
    @lazy
    def fiveputr(self):
        """ Returns the exon regions which contain 5'UTR as list of features """
        return self.get_features("5PUTR")

    
    def get_coding_sequence(self):
        """ Warning: There are frame shift issues not handled here.
            Do not naively turn this into a protein - better to use existing databases """
        return self.get_sequence_from_features("CDS")

    def get_5putr_sequence(self):
        return self.get_sequence_from_features("5PUTR")

    def get_3putr_sequence(self):
        return self.get_sequence_from_features("3PUTR")

    def get_intron_ivs(self):
        """
        Purpose: Get list of intron intervals for the the transcript
        Output: A list of HTSeq Genomic Interval instances for all introns in the transcript (ordered according to strand)
        """
        intron_ivs = []
        previous_exon = None
        for exon in self.get_features("exon"): # This is in stranded order
            if previous_exon:
                # HTSeq ends are 1 past the last base of the sequence.
                # Thus for touching sequences like exons/introns, first_seq.end = second_seq.start
                intron_start = previous_exon.iv.end_d
                intron_end = exon.iv.start_d
                intron_length = abs(intron_end - intron_start)
                intron = GenomicInterval_from_directional(exon.iv.chrom, intron_start, intron_length, exon.iv.strand)
                intron_ivs.append(intron)
            previous_exon = exon
        return intron_ivs
    
    def get_intron_sequences(self):
        """
        Get list of intron sequences for transcript
        Output: List of sequences (intron order and sequences are 5' to 3')
        """
        list_of_intron_intervals = self.get_intron_ivs()
        intron_sequences = []
        for intron in list_of_intron_intervals:
            intron_sequences.append(self.reference.get_sequence_from_iv(intron))
        return intron_sequences
    
    
    def get_genomic_position(self, pos_on_transcript):
        """
        Converts 0-based position on a transcript into 0-based position on the chromosome
        Arguments -- position relative to 5' end of transcript (int) 
        Returns -- position on chromosome (int)
        """
        #logging.debug("Searching %s for %d", self.get_id(), pos_on_transcript)
        
        running_exon_length = 0
        previous_running_exon_length = 0
        #go through exons in order, adding to running_exon_length, until you find the one with the miR seed start position in it
        for exon in self.get_features("exon"):
            running_exon_length += exon.iv.length
            if pos_on_transcript < running_exon_length: #match start is in this exon
                #logging.debug("found the exon, %r running exon length is: %r", exon, running_exon_length)
                #logging.debug("previous_running_exon_length is %r:", previous_running_exon_length)
                #logging.debug("exon start_d and end_d are: %r, %r", exon.iv.start_d, exon.iv.end_d)
                genomic_pos_of_exon_start = exon.iv.start_d
                pos_on_this_exon = pos_on_transcript - previous_running_exon_length
                #logging.debug("position on this exon is %r", pos_on_this_exon)
                if exon.iv.strand == '+':
                    genomic_position_of_match_start = genomic_pos_of_exon_start + pos_on_this_exon
                elif exon.iv.strand == '-':
                    genomic_position_of_match_start = genomic_pos_of_exon_start - pos_on_this_exon
                else:
                    raise ValueError("strand must be + or -, not: %r" % exon.iv.strand)
                return genomic_position_of_match_start
            previous_running_exon_length += exon.iv.length

        raise NotOnTranscriptException("%s didn't contain %s" % (self.get_id(), pos_on_transcript))

    def __repr__(self):
        coding_str = " (coding)" if self.is_coding else ""
        return "Transcript %s%s: length %d" % (self.get_id(), coding_str, self.length)



from __future__ import print_function, absolute_import

import HTSeq
from deprecation import deprecated
from functools import reduce
import gzip
import json
from lazy import lazy
import operator
import os
from pyreference import settings
from pyreference.gene import Gene
from pyreference.mirna import MiRNA
from pyreference.pyreference_config import load_params_from_config
from pyreference.settings import BEST_REGION_TYPE_ORDER
from pyreference.settings import CHROM, START, END, STRAND
from pyreference.transcript import Transcript
from pyreference.utils.genomics_utils import get_unique_features_from_genomic_array_of_sets_iv, fasta_to_hash, \
    HTSeqInterval_to_feature_dict, reverse_complement
from pysam import FastaFile  # @UnresolvedImport
import six
import sys


def _load_gzip_json(gz_json_file_name, use_gzip_open=True):
    decompress_in_memory = not use_gzip_open
    if use_gzip_open:
        try:
            with gzip.open(gz_json_file_name, "rb") as f:
                json_bytes = f.read()
        except IOError as e:
            # We sometimes get [Errno 5] Input/output error using CIFS (SMB)
            print(e, file=sys.stderr)
            if e.errno == 5:
                decompress_in_memory = True

    if decompress_in_memory:
        with open(gz_json_file_name, "rb") as f:
            gzip_bytes = f.read()
            json_bytes = gzip.GzipFile(fileobj=six.StringIO(gzip_bytes)).read()

        if use_gzip_open:
            msg = "gzip.open failed, successfully fell back on in-memory decompression\n"
            msg += "Please set use_gzip_open=False in your settings to speed up load times."
            print(msg, file=sys.stderr)

    if six.PY2:
        json_str = json_bytes
    else:
        json_str = json_bytes.decode('ascii')
    data = json.loads(json_str)

    pyreference_json_version = data[settings.PYREFERENCE_JSON_VERSION_KEY]
    if settings.PYREFERENCE_JSON_VERSION != pyreference_json_version:
        params = {"version_key": settings.PYREFERENCE_JSON_VERSION_KEY,
                  "current_version": settings.PYREFERENCE_JSON_VERSION,
                  "json_version": pyreference_json_version,
                  "file_name": gz_json_file_name}
        msg = "PyReference with %(version_key)s %(current_version)d attempted to load '%(file_name)s' with %(version_key)s: %(json_version)d.\n" % params
        msg += "Please re-create with this version of pyreference_gff_to_json.py."
        raise ValueError(msg)

    return data


class Reference(object):
    def __init__(self, build=None, config=None, **kwargs):
        """ Construct a new reference object via:
        
            build - from pyreference config file (defaults to [global] default_build from config file) 
            config - config file (defaults to ~/pyreference.cfg)
            
            OR pass in the file names:
            
            genes_json
            trna_json
            genome_sequence_fasta
            mature_mir_sequence_fasta
            
            
            Any passed parameters will overwrite those from the config file 
            
            stranded - interval tests are stranded? (default True) 
            
            """

        # May not need to have config file if they passed in params 
        config_exception = None
        try:
            params = load_params_from_config(build=build, config=config)
        except OSError as e:
            config_exception = e
            params = {"build": build}

        # Set / Overwrite with non-null kwargs
        params.update({k: v for (k, v) in kwargs.items() if v is not None})
        self._genes_json = params.get("genes_json")
        self._trna_json = params.get("trna_json")
        self._genome_sequence_fasta = params.get("genome_sequence_fasta")
        self._mature_mir_sequence_fasta = params.get("mature_mir_sequence_fasta")
        self.use_gzip_open = params.get("use_gzip_open", True)
        self.stranded = params.get("stranded", True)

        # Need at least this
        if self._genes_json is None:
            if kwargs:
                six.raise_from(ValueError("No 'genes_json' in passed kwargs"), config_exception)
            raise config_exception

        # Store this so we can ask about config later
        self.build = params["build"]
        self._args = {"build": build, "config": config}
        self._build_params = params

    @lazy
    def _genes_dict(self):
        return _load_gzip_json(self._genes_json, self.use_gzip_open)

    def get_transcript_dict(self, transcript_id):
        transcripts_by_id = self._genes_dict["transcripts_by_id"]
        return transcripts_by_id[transcript_id]

    @lazy
    def genes(self):
        """ dict of {"gene_id" : Gene} """

        genes_by_id = self._genes_dict["genes_by_id"]
        genes = {}
        for gene_id in genes_by_id:
            genes[gene_id] = self.get_gene_by_id(gene_id)
        return genes

    @lazy
    def transcripts(self):
        """ dict of {"transcript_id" : Transcript} """

        # Implement in terms of genes so that it shares Transcript objects (with gene set)        
        transcripts = {}
        for g in six.itervalues(self.genes):
            for t in g.transcripts:
                transcripts[t.get_id()] = t
        return transcripts

    @lazy
    def protein_coding_genes(self):
        """Return dict of {gene_name: Gene} for protein coding genes"""
        genes_dict = {}
        for gene in self.genes_by_biotype["protein_coding"]:
            genes_dict[gene.name] = gene
        return genes_dict

    @lazy
    def genes_by_biotype(self):
        """ dict of {"biotype" : array_of_genes_biotype }
            This also includes 'tRNA' (from non-standard UCSC GTF) """
        gene_ids_by_biotype = self._genes_dict["gene_ids_by_biotype"]

        genes_by_biotype = {}
        for (biotype, gene_ids) in gene_ids_by_biotype.items():
            genes = []
            for gene_id in gene_ids:
                genes.append(self.get_gene_by_id(gene_id))

            genes_by_biotype[biotype] = genes

        return genes_by_biotype

    def get_gene_by_id(self, gene_id):
        genes_by_id = self._genes_dict["genes_by_id"]
        gene_dict = genes_by_id.get(gene_id)
        if gene_dict is None:
            msg = "No Gene found with ID=%s" % gene_id
            raise ValueError(msg)
        return Gene(self, gene_id, gene_dict)

    def get_transcript_by_id(self, transcript_id):
        transcript_dict = self.get_transcript_dict(transcript_id)
        if transcript_dict is None:
            msg = "No Transcript found with ID=%s" % transcript_dict
            raise ValueError(msg)
        return Transcript(self, transcript_id, transcript_dict)

    def get_gene_by_name(self, gene_name):
        gene_id_by_name = self._genes_dict["gene_id_by_name"]
        gene_id = gene_id_by_name.get(gene_name)
        if gene_id is None:
            msg = "No Gene found with Name=%s" % gene_name
            raise ValueError(msg)
        return self.get_gene_by_id(gene_id)

    # @deprecated(details="Use get_gene_by_id")
    def get_gene(self, gene_id):
        return self.get_gene_by_id(gene_id)

    # @deprecated(details="Use get_transcript_by_id")
    def get_transcript(self, transcript_id):
        return self.get_transcript_by_id(transcript_id)

    def __getitem__(self, gene_ids):
        return self.get_genes_by_id(gene_ids)

    def get_genes_by_id(self, gene_ids):
        genes_subset = []
        for gene_id in gene_ids:
            genes_subset.append(self.get_gene_by_id(gene_id))
        return genes_subset

    def get_genes_by_name(self, gene_names):
        genes_subset = []
        for gene_name in gene_names:
            genes_subset.append(self.get_gene_by_name(gene_name))
        return genes_subset

    @lazy
    def mirna_mature(self):
        """Returns dict of { miRNA name: mature miR RNA sequence }"""
        if not self._mature_mir_sequence_fasta:
            msg = "Asked for miRNA mature sequence but not provided with a 'mature_mir_sequence_fasta' argument"
            raise RuntimeError(msg)
        return fasta_to_hash(self._mature_mir_sequence_fasta)

    def get_mirna(self, mirna_name):
        """Returns an MiRNA class instance for the mirna_name"""
        mirna = MiRNA(mirna_name, self.mirna_mature)
        return mirna

    @lazy
    def genome(self):
        if not self._genome_sequence_fasta:
            raise IOError("Asked for sequence but no genome_sequence_fasta provided.")
        if not os.path.exists(self._genome_sequence_fasta):
            raise IOError("Genome sequence file '%s' not found" % self.reference_fasta)

        return FastaFile(self._genome_sequence_fasta)

    def get_sequence_from_iv(self, iv, upper_case=True):
        feature_dict = HTSeqInterval_to_feature_dict(iv)
        return self.get_sequence_from_feature(feature_dict, upper_case=upper_case)

    def get_sequence_from_feature(self, feature_dict, upper_case=True):
        """Repetitive regions are sometimes represented as lower case.
            If upper_case=True, return the sequence as upper case (Default).
            If false, do not convert case, i.e retain lower case where it was present."""

        chrom = str(feature_dict[CHROM])
        start = feature_dict[START]
        end = feature_dict[END]
        strand = str(feature_dict[STRAND])
        seq = self.genome.fetch(reference=chrom,
                                start=start,
                                end=end)
        if strand == '-':
            seq = reverse_complement(seq)

        if upper_case:
            seq = seq.upper()
        return seq

    def get_sequence_from_features(self, features):
        sequences = []
        for feature in features:
            sequences.append(self.get_sequence_from_feature(feature))

        if sequences:
            sequence = reduce(operator.add, sequences)
        else:
            sequence = ""
        return sequence

    ###########################
    # Retrieve gene/transcript info by interval
    # Done by building a HTSeq GenomicArrayOfSets

    @lazy
    def genomic_transcripts(self):
        """ GenomicArrayOfSets containing transcripts - used for lookups """
        transcripts_gas = HTSeq.GenomicArrayOfSets("auto", self.stranded)
        for t in self.transcripts.values():
            transcripts_gas[t.iv] += t

        return transcripts_gas

    def get_transcripts_in_iv(self, iv):
        """Returns: list of transcripts in genomic interval"""
        transcripts = get_unique_features_from_genomic_array_of_sets_iv(self.genomic_transcripts, iv)
        return list(transcripts)

    def get_transcript_ids(self, iv):
        return [feature.get_id() for feature in self.get_transcripts_in_iv(iv)]

    def get_gene_names_array(self, iv):
        return list(set([t.get_gene_id() for t in self.get_transcripts_in_iv(iv)]))

    def get_gene_names(self, interval):
        """Returns a string of gene names"""
        gene_names = self.get_gene_names_array(interval)
        return " ".join(gene_names)

    ## Regions
    @lazy
    def regions(self):
        regions = HTSeq.GenomicArray("auto", stranded=self.stranded, typecode='O')

        # GenomicArray can only store 1 region for an interval, so go through 
        # and overwrite    
        for t in six.itervalues(self.transcripts):
            regions[t.iv] = "intron"

        for t in six.itervalues(self.transcripts):
            if not t.is_coding:
                for feature in t.get_features("exon"):
                    regions[feature.iv] = "non coding"

        for t in six.itervalues(self.transcripts):
            if t.is_coding:
                for utr in ["5PUTR", "3PUTR"]:
                    for feature in t.get_features(utr):
                        regions[feature.iv] = utr

        for t in six.itervalues(self.transcripts):
            if t.is_coding:
                for feature in t.get_features("CDS"):
                    regions[feature.iv] = "coding"

        return regions

    def get_regions_array(self, iv):
        """Returns: list of region types (str) in the interval"""
        _ = self.transcripts
        regions = []
        for (_, region_name) in self.regions[iv].steps():
            if region_name:
                regions.append(region_name)
        return regions

    def get_region_names(self, iv):
        region_names = set(self.get_regions_array(iv))
        return " ".join(region_names)

    def get_best_region(self, iv):
        """ Returns "best" region - if multiple pick according to order of BEST_REGION_TYPE_ORDER """

        region_names = set(self.get_regions_array(iv))
        region = None
        for r in BEST_REGION_TYPE_ORDER:
            if r in region_names:
                region = r
                break
        return region

    @deprecated(details="Use get_best_region()")
    def get_region(self, iv):
        return self.get_best_region(iv)

    @lazy
    def has_chr(self):
        transcripts_by_id = self._genes_dict["transcripts_by_id"]
        some_transcript = six.next(six.itervalues(transcripts_by_id))
        chrom = some_transcript["chr"]
        return chrom.startswith("chr")

    def __repr__(self):
        return "PyReference (%s)" % self.build

    @lazy
    def config(self):
        params = {"build": self.build,
                  "args": self._args,
                  "reference_gtf": self._genes_dict["reference_gtf"],
                  "build_params": self._build_params.copy()}
        return params



from __future__ import print_function, absolute_import

import abc
from lazy import lazy

from pyreference.utils.genomics_utils import iv_from_pos_range, \
    iv_from_pos_directional_before_after, dict_to_iv


class GenomicRegion(object):
    """ Base class for both Gene and Transcript """
    def __init__(self, reference, accession_id, data_dict):
        self.reference = reference
        self.accession_id = accession_id
        self._dict = data_dict
    
    def get_id(self):
        return self.accession_id

    @property
    def biotype(self):
        return '/'.join(sorted(self.get_biotypes()))

    def get_biotypes(self):
        return self._dict["biotype"]
    
    @lazy
    def iv(self):
        return dict_to_iv(self._dict)

    @lazy    
    def tss(self):
        """ (Representative) Transcript Start Site
            This is NOT the most 5' position (use iv.start_d_as_pos for that) """
        transcript_iv = self.get_representative_transcript().iv
        return transcript_iv.start_d_as_pos

    def get_promoter_iv(self, promoter_range=1000):
        return iv_from_pos_range(self.tss, promoter_range)

    def get_promoter_sequence(self, promoter_range=1000):
        iv = self.get_promoter_iv(promoter_range)
        return self.reference.get_sequence_from_iv(iv)


    def get_promoter_iv_custom_range(self, upstream_distance, downstream_distance):
        """Get any interval surrounding TSS
        Note: total length of interval = upstream_distance + downstream_distance (The TSS base is included in downstream_distance)"""
        return iv_from_pos_directional_before_after(self.tss, upstream_distance, downstream_distance)
    
    def get_promoter_sequence_custom_range(self, upstream_distance, downstream_distance):
        iv = self.get_promoter_iv_custom_range(upstream_distance, downstream_distance)
        return self.reference.get_sequence_from_iv(iv)

    @abc.abstractmethod
    def get_representative_transcript(self):
        pass



"""
Created on 22Jan.,2018

@author: dlawrence
"""

from Bio.Seq import Seq
from Bio import SeqIO
import HTSeq

try:
    from sys import intern
except (ImportError,AttributeError):
    pass
    
from pyreference.settings import CHROM, START, END, STRAND


def HTSeqInterval_to_feature_dict(iv):
    return {CHROM : iv.chrom, START : iv.start, END : iv.end, STRAND : iv.strand}

def dict_to_iv(data):
    chrom = str(data[CHROM])
    start = data[START]
    end = data[END]
    strand = str(data[STRAND])
    return HTSeq.GenomicInterval(chrom, start, end, strand)


def iv_from_pos_range(g_pos, range_length):
    """p = Genomic position.
    Returns iv 'range_length' bp upstream and 'range_length' downstream of position p"""
    return HTSeq.GenomicInterval( g_pos.chrom, g_pos.pos - range_length, g_pos.pos + range_length, g_pos.strand)

def iv_from_pos_directional_before_after(g_pos, upstream_length, downstream_length):
    """Note: The g_pos base is assumed to be included in downstream_length
    e.g upstream_length=100, downstream_length=100 has total length=200 bp
    Check that this function does what you're expecting - setting upstream length to 0 will change the position of start_d on neg strand..."""
    if g_pos.strand == '+':
        start = g_pos.pos - upstream_length
        end = g_pos.pos + downstream_length
    elif g_pos.strand == '-':
        start = g_pos.pos - downstream_length
        end = g_pos.pos + upstream_length
    else:
        raise ValueError("Unknown strand in genomic position %s" % g_pos)

    return HTSeq.GenomicInterval( g_pos.chrom, start, end, g_pos.strand)

def GenomicInterval_from_directional( chrom, start_d, length, strand="." ):
    """ Fix bug in HTSeq:
        HTSeq.GenomicInterval_from_directional throws 'str' object has no attribute 'se' """
    strand = intern( strand )
    if strand != "-":
        return HTSeq.GenomicInterval( chrom, start_d, start_d+length, strand )
    else:
        return HTSeq.GenomicInterval( chrom, start_d-length+1, start_d+1, strand )


def get_unique_features_from_genomic_array_of_sets_iv(genomic_array_of_sets, iv):
    """Collapse genomic array of sets into a unique list of the values in that region"""
    all_values = set()
    for iv, values_in_iv in genomic_array_of_sets[iv].steps():
        all_values.update(values_in_iv)
    return all_values


# HTSeq documentation says end = the end of the interval. Following Python convention for ranges, this
# is *ONE MORE* than the coordinate of the last base that is considered part of the sequence.
def last_base(iv):
    last_base_pos = iv.end_as_pos
    last_base_pos.pos -= 1
    return last_base_pos


def opposite_strand(strand):
    opposites = {"+" : "-",
                 "-" : "+"}
    o = opposites.get(strand)
    if o is None:
        raise ValueError("Unknown strand '%s'" % strand)
    return o


def format_chrom(chrom, want_chr):
    """ Pass in a chromosome (unknown format), return in your format
        @param chrom: chromosome ID (eg 1 or 'chr1')
        @param want_chr: Boolean - whether you want "chr" at the beginning of chrom
        @return: "chr1" or "1" (for want_chr True/False) 
    """
    if want_chr:   
        if chrom.startswith("chr"):
            return chrom
        else:
            return "chr%s" % chrom
    else:
        return chrom.replace("chr", "")


def fasta_to_hash(fasta):
    indexed_fasta = {}
    for record in SeqIO.parse(fasta, "fasta"):
        indexed_fasta[record.id] = str(record.seq)
    return indexed_fasta


def reverse_complement(dna_sequence):
    seq = Seq(dna_sequence)
    return str(seq.reverse_complement())



"""
Created on 24Jan.,2018

@author: dlawrence
"""
import configargparse
from pyreference import Reference


class ReferenceArgumentParser(configargparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super(ReferenceArgumentParser, self).__init__(*args, **kwargs)
        self.add("--build", env_var="BUILD", help='Use [build] section of config file.')
        self.add("--config", is_config_file=True, help='Config file used for [build] section')
        self.add("--genes_json", help='gtf.json.gz file created by pyreference_gff_to_json')
        self.add("--trna_json", help='gtf.json.gz file created by pyreference_gff_to_json')
        self.add("--mature_mir_sequence_fasta", help='MiRNA Fasta') 
        self.add("--genome_sequence_fasta", help='Genome Sequence Fasta')
        self.set_defaults(stranded=True)
        self.add('--stranded', dest='stranded', action='store_true')
        self.add('--unstranded', dest='stranded', action='store_false')


    def parse_args(self):
        """ get args from command line, adding 'reference' field set to PyReference instance """
        args = super(ReferenceArgumentParser, self).parse_args()
        args.reference = Reference(**args.__dict__)
        return args


    




"""
Created on 22Jan.,2018

@author: dlawrence
"""
from hashlib import md5
import os
import six


try:
    from pathlib import Path #@UnresolvedImport
except (ImportError,AttributeError):
    from pathlib2 import Path #@UnresolvedImport

def name_from_file_name(file_name):
    return Path(file_name).name

def stem_from_file_name(file_name):
    return Path(file_name).stem

def mk_path(path):
    if path and not os.path.exists(path):
        os.makedirs(path)

def mk_path_for_file(f):
    mk_path(os.path.dirname(f))
    
def file_or_file_name(f, mode='r'):
    if isinstance(f, six.string_types):
        if 'w' in mode: # Create path if writing
            mk_path_for_file(f)
            
        return open(f, mode)
    elif hasattr(f, 'read'):
        return f # Already a File object
    else:
        raise ValueError("'%s' (%s) not a file or string" % (f, type(f)))

def file_md5sum(filename):
    m = md5()
    with open(filename, "rb") as f:
        m.update(f.read())
    return m.hexdigest()


#!/usr/bin/env python

from __future__ import print_function, absolute_import

import HTSeq
import abc
import gzip
import json
import logging
import os
from argparse import ArgumentParser
from collections import defaultdict, Counter

from pyreference.settings import CHROM, START, END, STRAND, IS_CODING, \
    PYREFERENCE_JSON_VERSION_KEY, PYREFERENCE_JSON_VERSION
from pyreference.utils.file_utils import name_from_file_name, file_md5sum


class GFFParser(abc.ABC):
    CODING_FEATURES = {"CDS", "start_codon", "stop_codon"}
    FEATURE_ALLOW_LIST = {}
    FEATURE_IGNORE_LIST = {"biological_region", "chromosome", "region", "scaffold", "supercontig"}

    def __init__(self, filename, discard_contigs_with_underscores=True):
        self.filename = filename
        self.discard_contigs_with_underscores = discard_contigs_with_underscores

        self.discarded_contigs = Counter()
        self.genes_by_id = {}
        self.transcripts_by_id = {}
        self.gene_id_by_name = {}
        # Store CDS in separate dict as we don't need to write as JSON
        self.transcript_cds_by_id = {}

    @abc.abstractmethod
    def handle_feature(self, feature):
        pass

    def parse(self):
        for feature in HTSeq.GFF_Reader(self.filename):
            if self.FEATURE_ALLOW_LIST and feature.type not in self.FEATURE_ALLOW_LIST:
                continue
            if feature.type in self.FEATURE_IGNORE_LIST:
                continue

            try:
                chrom = feature.iv.chrom
                if self.discard_contigs_with_underscores and not chrom.startswith("NC_") and "_" in chrom:
                    self.discarded_contigs[chrom] += 1
                    continue
                self.handle_feature(feature)
            except Exception as e:
                print("Could not parse '%s': %s" % (feature.get_gff_line(), e))
                raise e

    def finish(self):
        self._add_coding_and_utr_features()

        if self.discarded_contigs:
            print("Discarded contigs: %s" % self.discarded_contigs)


    @staticmethod
    def _create_gene(gene_name, feature):
        biotypes = set()

        gene = {
            "name": gene_name,
            "transcripts": set(),
            "biotype": biotypes,
            CHROM: feature.iv.chrom,
            START: feature.iv.start,
            END: feature.iv.end,
            STRAND: feature.iv.strand
        }

        # Attempt to get some biotypes in there if available
        if feature.type == "gene":
            gene_version = feature.attr.get("version")
            biotype = feature.attr.get("biotype")
            description = feature.attr.get("description")
            if description:
                gene["description"] = description
        else:
            gene_version = feature.attr.get("gene_version")
            biotype = feature.attr.get("gene_biotype")

        if biotype:
            biotypes.add(biotype)

        if gene_version:
            gene["version"] = int(gene_version)
        return gene

    @staticmethod
    def _create_transcript(feature):
        return {
            "features_by_type": defaultdict(list),
            "biotype": set(),
            CHROM: feature.iv.chrom,
            START: feature.iv.start,
            END: feature.iv.end,
            STRAND: feature.iv.strand,
            IS_CODING: 0
        }

    @staticmethod
    def _store_other_chrom(data, feature):
        other_chroms = data.get("other_chroms", set())
        other_chroms.add(feature.iv.chrom)
        data["other_chroms"] = other_chroms

    @staticmethod
    def _get_biotype_from_transcript_id(transcript_id):
        biotypes_by_transcript_id_start = {"NM_": "protein_coding", "NR_": "non_coding"}
        for (start, biotype) in biotypes_by_transcript_id_start.items():
            if transcript_id.startswith(start):
                return biotype

        if "tRNA" in transcript_id:
            return "tRNA"
        return "N/A"

    def _add_transcript_data(self, transcript_id, transcript, feature):
        if feature.iv.chrom != transcript[CHROM]:
            self._store_other_chrom(transcript, feature)
            return

        feature_dict = {START: feature.iv.start,
                        END: feature.iv.end}
        if feature.type == "cDNA_match":
            target = feature.attr.get("Target")
            t_cols = target.split()
            feature_dict["cdna_start"] = int(t_cols[1])
            feature_dict["cdna_end"] = int(t_cols[2])
            if len(t_cols) == 4 and t_cols[3] != '+':  # Default is '+', so only store '-'
                feature_dict["cdna_strand"] = t_cols[3]
            gap = feature.attr.get("Gap")
            if gap:
                feature_dict["gap"] = gap

        transcript["features_by_type"][feature.type].append(feature_dict)
        if feature.type in self.CODING_FEATURES:
            cds_extent = self.transcript_cds_by_id.get(transcript_id)
            if cds_extent is None:
                cds_extent = {START: feature.iv.start,
                              END: feature.iv.end}
                self.transcript_cds_by_id[transcript_id] = cds_extent
            else:
                cds_extent[START] = min(cds_extent[START], feature.iv.start)
                cds_extent[END] = max(cds_extent[END], feature.iv.end)

    def _add_coding_and_utr_features(self):
        """ Add 5PUTR/3PUTR features to coding transcripts

            Ensembl GTFs have 'five_prime_UTR' features (similar to CDS etc) but we make this for GFFs that
            don't have those features
        """

        for transcript_id, transcript in self.transcripts_by_id.items():
            cds_extent = self.transcript_cds_by_id.get(transcript_id)
            if cds_extent:
                transcript[IS_CODING] = 1
                features_by_type = transcript["features_by_type"]

                (left, right) = ("5PUTR", "3PUTR")
                if transcript[STRAND] == '-':  # Switch
                    (left, right) = (right, left)

                cds_min = cds_extent[START]
                cds_max = cds_extent[END]

                transcript["cds_start"] = cds_min
                transcript["cds_end"] = cds_max

                # exon is in stranded order
                for exon in features_by_type["exon"]:
                    exon_start = exon[START]
                    exon_end = exon[END]

                    if exon_start < cds_min:
                        end_non_coding = min(cds_min, exon_end)
                        utr_feature = {START: exon_start,
                                       END: end_non_coding}
                        features_by_type[left].append(utr_feature)

                    if exon_end > cds_max:
                        start_non_coding = max(cds_max, exon_start)
                        utr_feature = {START: start_non_coding,
                                       END: exon_end}
                        features_by_type[right].append(utr_feature)

    def get_data(self):
        self.parse()
        self.finish()

        gene_ids_by_biotype = defaultdict(set)
        for gene_id, gene in self.genes_by_id.items():
            for biotype in gene["biotype"]:
                gene_ids_by_biotype[biotype].add(gene_id)

        return {
            PYREFERENCE_JSON_VERSION_KEY: PYREFERENCE_JSON_VERSION,
            "reference_gtf": {"path": os.path.abspath(self.filename),
                              "md5sum": file_md5sum(self.filename)},
            "genes_by_id": self.genes_by_id,
            "transcripts_by_id": self.transcripts_by_id,
            "gene_id_by_name": self.gene_id_by_name,
            "gene_ids_by_biotype": gene_ids_by_biotype,
        }


class GTFParser(GFFParser):
    """ GTF (GFF2) - used by Ensembl, @see http://gmod.org/wiki/GFF2

        GFF2 only has 2 levels of feature hierarchy, so we have to build or 3 levels of gene/transcript/exons ourselves
    """
    GTF_TRANSCRIPTS_DATA = GFFParser.CODING_FEATURES | {"exon"}
    FEATURE_ALLOW_LIST = GTF_TRANSCRIPTS_DATA | {"gene"}

    def __init__(self, *args, **kwargs):
        super(GTFParser, self).__init__(*args, **kwargs)

    def handle_feature(self, feature):
        gene_id = feature.attr["gene_id"]
        # Non mandatory - Ensembl doesn't have on some RNAs
        gene_name = None
        if feature.type == "gene":
            gene_name = feature.attr.get("Name")
        else:
            gene_name = feature.attr.get("gene_name")
        if gene_name:
            self.gene_id_by_name[gene_name] = gene_id  # Shouldn't be dupes per file

        gene = self.genes_by_id.get(gene_id)
        if gene is None:
            gene = self._create_gene(gene_name, feature)
            self.genes_by_id[gene_id] = gene
        else:
            self._update_extents(gene, feature)

        transcript_id = feature.attr.get("transcript_id")
        transcript_version = feature.attr.get("transcript_version")
        if transcript_version:
            transcript_id += "." + transcript_version

        if transcript_id:
            gene["transcripts"].add(transcript_id)
            transcript = self.transcripts_by_id.get(transcript_id)
            if transcript is None:
                transcript = self._create_transcript(feature)
                self.transcripts_by_id[transcript_id] = transcript
            else:
                self._update_extents(transcript, feature)

            # No need to store chrom/strand for each feature, will use transcript
            if feature.type in self.GTF_TRANSCRIPTS_DATA:
                self._add_transcript_data(transcript_id, transcript, feature)

            biotype = feature.attr.get("gene_biotype")
            if biotype is None:
                biotype = self._get_biotype_from_transcript_id(transcript_id)

            if biotype:
                gene["biotype"].add(biotype)
                transcript["biotype"].add(biotype)

    @staticmethod
    def _update_extents(genomic_region_dict, feature):
        if feature.iv.chrom == genomic_region_dict[CHROM]:
            start = genomic_region_dict[START]
            if feature.iv.start < start:
                genomic_region_dict[START] = feature.iv.start

            end = genomic_region_dict[END]
            if feature.iv.end > end:
                genomic_region_dict[END] = feature.iv.end
        else:
            self._store_other_chrom(genomic_region_dict, feature)


class GFF3Parser(GFFParser):
    """ GFF3 - Used by RefSeq, @see https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md

        GFF3 support arbitrary hierarchy

    """

    GFF3_GENES = {"gene", "pseudogene"}
    GFF3_TRANSCRIPTS_DATA = {"exon", "CDS", "cDNA_match", "five_prime_UTR", "three_prime_UTR"}

    def __init__(self, *args, **kwargs):
        super(GFF3Parser, self).__init__(*args, **kwargs)
        self.gene_id_by_feature_id = defaultdict()
        self.transcript_id_by_feature_id = defaultdict()

    def handle_feature(self, feature):
        parent_id = feature.attr.get("Parent")
        # Genes never have parents
        # RefSeq genes are always one of GFF3_GENES, Ensembl has lots of different types (lincRNA_gene etc)
        # Ensembl treats pseudogene as a transcript (has parent)
        if parent_id is None and (feature.type in self.GFF3_GENES or "gene_id" in feature.attr):
            gene_id = feature.attr.get("gene_id")
            dbxref = self._get_dbxref(feature)
            if not gene_id:
                gene_id = dbxref.get("GeneID")
            if not gene_id:
                raise ValueError("Could not obtain 'gene_id', tried 'gene_id' and 'Dbxref[GeneID]'")

            gene_name = feature.attr.get("Name")
            # Gene can have multiple loci, thus entries in GFF, keep original so all transcripts are added
            gene = self.genes_by_id.get(gene_id)
            if gene is None:
                gene = self._create_gene(gene_name, feature)
                # If a gene already exists - then need to merge it...
                self.genes_by_id[gene_id] = gene

            hgnc = dbxref.get("HGNC")
            if hgnc:
                gene["HGNC"] = hgnc

            if gene_name:
                self.gene_id_by_name[gene_name] = gene_id
            self.gene_id_by_feature_id[feature.attr["ID"]] = gene_id
        else:
            if feature.type in self.GFF3_TRANSCRIPTS_DATA:
                if feature.type == 'cDNA_match':
                    target = feature.attr["Target"]
                    transcript_id = target.split()[0]
                else:
                    # Some exons etc may be for miRNAs that have no transcript ID, so skip those (won't have parent)
                    if parent_id:
                        transcript_id = self.transcript_id_by_feature_id.get(parent_id)
                    else:
                        logging.warning("Transcript data has no parent: %s" % feature.get_gff_line())
                        transcript_id = None

                if transcript_id:
                    transcript = self.transcripts_by_id[transcript_id]
                    self._handle_transcript_data(transcript_id, transcript, feature)
            else:
                # There are so many different transcript ontology terms just taking everything that
                # has a transcript_id and is child of gene (ie skip miRNA etc that is child of primary_transcript)
                transcript_id = feature.attr.get("transcript_id")
                if transcript_id:
                    transcript_version = feature.attr.get("version")
                    if transcript_version:
                        transcript_id += "." + transcript_version
                    assert parent_id is not None
                    gene_id = self.gene_id_by_feature_id.get(parent_id)
                    if not gene_id:
                        raise ValueError("Don't know how to handle feature type %s (not child of gene)" % feature.type)
                    gene = self.genes_by_id[gene_id]
                    self._handle_transcript(gene, transcript_id, feature)

    @staticmethod
    def _get_dbxref(feature):
        """ RefSeq stores attribute with more keys, eg: 'Dbxref=GeneID:7840,HGNC:HGNC:428,MIM:606844' """
        dbxref = {}
        dbxref_str = feature.attr.get("Dbxref")
        if dbxref_str:
            dbxref = dict(d.split(":", 1) for d in dbxref_str.split(","))
        return dbxref

    def _handle_transcript(self, gene, transcript_id, feature):
        """ Sometimes we can get multiple transcripts in the same file - just taking 1st """
        if transcript_id not in self.transcripts_by_id:
            # print("_handle_transcript(%s, %s)" % (gene, feature))
            transcript = self._create_transcript(feature)
            biotype = self._get_biotype_from_transcript_id(transcript_id)
            gene["transcripts"].add(transcript_id)
            gene["biotype"].add(biotype)
            transcript["biotype"].add(biotype)
            partial = feature.attr.get("partial")
            if partial:
                transcript["partial"] = 1
            self.transcripts_by_id[transcript_id] = transcript
        self.transcript_id_by_feature_id[feature.attr["ID"]] = transcript_id

    def _handle_transcript_data(self, transcript_id, transcript, feature):
        self._add_transcript_data(transcript_id, transcript, feature)


def handle_args():
    parser = ArgumentParser(description='Build a json.gz file for pyreference')
    parser.add_argument("--discard-contigs-with-underscores", action='store_true', default=True)
    parser.add_argument('--url', help='URL (source of GFF) to store in "reference_gtf.url"')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--gtf', help='GTF (Gene Transfer Format) filename')
    group.add_argument('--gff3', help='GFF3 (Gene Feature Format) filename')
    args = parser.parse_args()
    if not (args.gtf or args.gff3):
        parser.error("You must specify either --gtf or --gff3")
    return args


def parser_factory(gtf=None, gff3=None, discard_contigs_with_underscores=True):
    if gtf:
        parser = GTFParser(gtf, discard_contigs_with_underscores)
    else:
        parser = GFF3Parser(gff3, discard_contigs_with_underscores)
    return parser


class SortedSetEncoder(json.JSONEncoder):
    """ Dump set as list, from: https://stackoverflow.com/a/8230505/295724 """

    def default(self, obj):
        if isinstance(obj, set):
            return list(sorted(obj))
        return json.JSONEncoder.default(self, obj)


def main():
    args = handle_args()
    parser = parser_factory(args.gtf, args.gff3,
                            discard_contigs_with_underscores=args.discard_contigs_with_underscores)
    data = parser.get_data()
    if args.url:
        data["reference_gtf"]["url"] = args.url

    genes_json_gz = name_from_file_name(parser.filename) + ".json.gz"
    with gzip.open(genes_json_gz, 'w') as outfile:
        json_str = json.dumps(data, cls=SortedSetEncoder, sort_keys=True)  # Sort so diffs work
        outfile.write(json_str.encode('ascii'))

    print("Wrote:", genes_json_gz)


if __name__ == '__main__':
    main()



"""
Created on 22Jan.,2018

TODO: Link to wiki page about config file format

@author: dlawrence
"""

import os
from six import raise_from
from six.moves import configparser  # @UnresolvedImport

try:
    from ConfigParser import SafeConfigParser as ConfigParser
except ImportError:
    from configparser import ConfigParser


def load_params_from_config(build=None, config=None):
    DEFAULT_PYREFERENCE_CFG = "~/pyreference.cfg"

    if config is None:
        config = os.path.expanduser(DEFAULT_PYREFERENCE_CFG)
        if not os.path.exists(config):
            msg = "No default config file: '%s' found. Please create one or pass another with the 'pyreference_cfg' parameter." % config
            raise OSError(msg)
    else:
        if not os.path.exists(config):
            msg = "Passed config file '%s' not found." % config
            raise OSError(msg)

    GLOBAL_FLAGS = ["use_gzip_open", "stranded"]
    params = {}

    defaults = {'genes_json': None,
                'trna_json': None,
                'mature_mir_sequence_fasta': None,
                'genome_sequence_fasta': None, }
    cfg = ConfigParser(allow_no_value=True, defaults=defaults)
    cfg.read(config)

    if build is None:
        try:
            build = cfg.get("global", "default_build")
        except configparser.NoOptionError as noe:
            msg = "No build parameter passed, and no default_build set in [global] section of config file '%s'" % config
            raise_from(configparser.NoOptionError(msg), noe)

    if not cfg.has_section(build):
        msg_params = {"build": build, "config": config}
        msg = "Build='%(build)s', no section [%(build)s] in config file '%(config)s" % msg_params
        raise ValueError(msg)

    for f in GLOBAL_FLAGS:
        try:
            params[f] = cfg.getboolean("global", f)
        except configparser.NoOptionError as noe:
            pass

    params["build"] = build
    params["config"] = config

    for k in defaults.keys():
        params[k] = cfg.get(build, k)

    return params



"""
Created on 23Jan.,2018

@author: dlawrence
"""


# Change this when you introduce breaking changes
PYREFERENCE_JSON_VERSION = 5
PYREFERENCE_JSON_VERSION_KEY = "pyreference_json_version"

# Keys used in dictionary (serialized to JSON)
CHROM = "chr"
START = "start"
END = "stop"
STRAND = "strand"

# Other 
IS_CODING = "is_coding"


BEST_REGION_TYPE_ORDER = ["coding", "5PUTR", "3PUTR", "non coding", "intron"]




"""
Unit tests for sacgf.genomics.reference.Reference

Created on Jun 20, 2012

@author: dlawrence
"""

from __future__ import print_function, absolute_import

import HTSeq
from inspect import getsourcefile
import os
from os.path import abspath
import six
import unittest

from pyreference import Reference


class Test(unittest.TestCase):
    def setUp(self):
        this_file_dir = os.path.dirname(abspath(getsourcefile(lambda: 0)))
        reference_dir = os.path.join(this_file_dir, "reference")

        genes_json = os.path.join(reference_dir, "hg19_chrY_300kb_genes.gtf.json.gz")
        genome_sequence_fasta = os.path.join(reference_dir, "hg19_chrY_300kb.fa")
        mature_mir_sequence_fasta = os.path.join(reference_dir, "mature_200ab_only.fa")

        self.reference = Reference(genes_json=genes_json,
                                   genome_sequence_fasta=genome_sequence_fasta,
                                   mature_mir_sequence_fasta=mature_mir_sequence_fasta)

    def tearDown(self):
        pass

    def test_regions(self):
        transcript_id = "NM_018390_2"
        transcript = self.reference.transcripts[transcript_id]

        for cds in transcript.get_features("CDS"):
            coding_regions = self.reference.get_region_names(cds.iv)
            self.assertEqual("coding", coding_regions)

        gene_regions = self.reference.get_regions_array(transcript.iv)
        self.assertTrue("intron" in gene_regions)
        self.assertTrue("coding" in gene_regions)
        self.assertTrue("5PUTR" in gene_regions)
        self.assertTrue("3PUTR" in gene_regions)

    def test_region_loading(self):
        # TODO: This call fails! Fix it!
        #        self.assertTrue(self.reference.regions is not None, "Fix me!")

        # Load AFTER call to transcripts works        
        self.reference.transcripts
        self.assertTrue(self.reference.regions is not None)

    def test_genes(self):
        test_cases = {  # Yes, I actually checked these in IGV
            "PLCXD1": {"transcript_id": "NM_018390_2",
                       "3p_utr": "CGGGACCCTT"},
            "PPP2R3B": {"transcript_id": "NM_013239",  # -'ve strand
                        "3p_utr": "CGCCGCCCGC"},
        }

        for test in six.itervalues(test_cases):
            transcript = self.reference.transcripts[test["transcript_id"]]

            utr = str(transcript.get_3putr_sequence()).upper()
            self.assertTrue(utr.startswith(test["3p_utr"]), "%s should start with %s" % (utr, test["3p_utr"]))

            m_rna = transcript.get_transcript_sequence()
            self.assertTrue(m_rna.find(test["3p_utr"]) > 1)

    def test_get_transcript_length(self):
        transcript_id = "NM_018390_2"
        transcript = self.reference.transcripts[transcript_id]

        seq = transcript.get_transcript_sequence()
        #        print "**** transcript_id = %s - sequence = %s" % (transcript_id, seq)
        sequence_length = len(seq)
        length = transcript.get_transcript_length()
        self.assertEqual(length, sequence_length)

    def test_mirna(self):
        miR = self.reference.get_mirna("hsa-miR-200a-3p")
        seed = miR.get_8mer_target()
        # print "seed = %s" % seed
        self.assertEqual(seed, "CAGTGTTA", "200a seed match")

        miR = self.reference.get_mirna("hsa-miR-200b-3p")
        seed = miR.get_8mer_target()
        # print "seed = %s" % seed
        self.assertEqual(seed, "CAGTATTA", "200b seed match")

    def test_promoter(self):
        tests = {"NM_018390_2": "CCGGGCAGCAGGGAAGATCT", "NM_013239": "CGCAGTGACGTGAACGCGGG"}

        for (tss_id, expected_promoter_sequence) in six.iteritems(tests):
            transcript = self.reference.transcripts[tss_id]
            promoter_sequence = str(transcript.get_promoter_sequence(10)).upper()
            # print "promoter sequence = %s" % promoter_sequence
            self.assertEqual(expected_promoter_sequence, promoter_sequence,
                             "%s strand promoter sequence" % transcript.iv.strand)

    def test_get_gene_names(self):
        intron = HTSeq.GenomicInterval("chrY", 144043, 144218, '+')
        gene_name = self.reference.get_gene_names(intron)
        self.assertEquals("PLCXD1", gene_name)

    def test_get_gene_region_names(self):
        intron = HTSeq.GenomicInterval("chrY", 144043, 144218, '+')
        region = self.reference.get_region_names(intron)
        self.assertEquals("intron", region)

    def test_gene_transcripts(self):
        plcxd1 = self.reference.get_gene("PLCXD1")
        expected_transcripts = ["NR_028057_2", "NM_018390_2"]  # chrY ones have _2

        plcxd1_transcript_ids = {pt.get_id() for pt in plcxd1.transcripts}
        message = "plcxd1 gene contains transcripts %s" % expected_transcripts
        for et_id in expected_transcripts:
            self.assertIn(et_id, plcxd1_transcript_ids, message)

    def test_has_chrom(self):
        self.assertTrue(self.reference.has_chr)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_name']
    unittest.main()



"""
Created on 19Jan.,2018

@author: dlawrence
"""

from lazy import lazy
from pyreference.genomic_region import GenomicRegion
from pyreference.transcript import Transcript
import sys


class Gene(GenomicRegion):
    """ Gene (which could contain multiple transcripts) """

    @property
    def name(self):
        return self.get_gene_name()

    def get_gene_name(self):
        return self._dict["name"]

    @lazy
    def transcripts(self):
        transcripts = []
        
        for transcript_id in self._dict["transcripts"]:
            td = self.reference.get_transcript_dict(transcript_id)  
            transcript = Transcript(self.reference, transcript_id, td, gene=self)
            transcripts.append(transcript)
        return transcripts


    @lazy
    def is_coding(self):
        return any(t.is_coding for t in self.transcripts)


    @lazy
    def representative_transcript(self):
        """ Returns longest coding transcript if gene is coding, otherwise longest transcript
            Sort transcript ID alphabetically if equal length """
        
        transcript = self.get_longest_coding_transcript()
        if transcript == None:
            transcript = self.get_longest_transcript()
        return transcript
    

    def get_representative_transcript(self):
        return self.representative_transcript
        

    def get_longest_coding_transcript(self):
        return self.get_longest_transcript(coding_only=True)

    
    def get_longest_transcript(self, coding_only=False):
        transcripts = self.transcripts
        if coding_only:
            transcripts = filter(lambda t : t.is_coding, transcripts)
        
        longest_transcript = None
        if transcripts:
            # We want the MAX length - and MIN ID, so sort by min but use maxint-length  
            # We also want NM_007041 (len 2209) over NM_001001976 (len 2209)
            # Which is annoyingly zero padded - so use smallest ID length, then only if equal do alpha sort 
            def min_transcript_key(t):
                return (sys.maxint - t.length, len(t.get_id()), t.get_id())

            longest_transcript = min(transcripts, key=min_transcript_key)
        return longest_transcript


    def __repr__(self):
        return "%s (%s) %d transcripts" % (self.get_gene_name(), self.accession_id, len(self.transcripts))
        


"""

ACRF Cancer Genomics Facility

Created on 23Jan.,2018

@see http://www.mirbase.org/

@author: dlawrence
"""

from Bio.Seq import Seq

def rna_to_dna(rna_sequence):
    rna = Seq(rna_sequence)
    return rna.back_transcribe()

def rna_to_target(rna_sequence):
    dna = rna_to_dna(rna_sequence)
    return str(dna.reverse_complement())

class MiRNA(object):
    def __init__(self, name, ref_sequences):
        self.name = name
        self.mature = ref_sequences[self.name]

    def get_8mer_target(self):
        return rna_to_target(self.mature[:8])

    # Nucleotides 4-15
    def get_central_paired_target(self):
        return rna_to_target(self.mature[3:15])

    def get_mature_dna(self):
        return str(rna_to_dna(self.mature))




from distutils.core import setup
from setuptools import find_packages

setup(name='pyreference',
      packages=find_packages(),
      version='0.6',
      description='Library for working with reference genomes',
      author='David Lawrence',
      author_email='davmlaw@gmail.com',
      url='https://github.com/SACGF/pyreference',
      keywords=['genomics', 'gtf', 'gff', 'genome', 'genes'],
      classifiers=[],
      install_requires=[
          'biopython',
          'configargparse',
          'deprecation',
          'HTSeq',
          'lazy',
          'pysam',
      ],
      python_requires='>=2.7, >=3.5',
      scripts=['bin/pyreference_gff_to_json.py'], )



"""
Created on 22Jan.,2018

@author: dlawrence
"""
import HTSeq
import os

from pyreference.utils.genomics_utils import format_chrom


def format_chrom_iterator(iterator, want_chr):
    for iv in iterator:
        iv.chrom = format_chrom(iv.chrom, want_chr)
        yield iv


def load_iv_iterator(file_name):
    suffix = os.path.splitext(file_name)[1].lower()
    iterator = None
    if suffix == ".bam":
        iterator = bam_iv_iterator(file_name)
    elif suffix == ".bed":
        # TODO:
        raise NotImplementedError("No bed reader yet")
    elif suffix == '.sam':
        iterator = sam_iv_iterator(file_name)
    elif suffix in ['.gff', '.gtf']:
        iterator = gff_iv_iterator(file_name)
    else:
        raise ValueError("Unknown input_file_type of " + suffix)
    return iterator

def chromosome_filter_iterator(chromosomes, iterator):
    for iv in iterator:
        if iv.chrom in chromosomes:
            yield iv

def bam_iv_iterator(bam_file):
    for aln in HTSeq.BAM_Reader(bam_file):
        if aln.aligned:
            yield aln.iv


def sam_iv_iterator(sam_file):
    for aln in HTSeq.SAM_Reader(sam_file):
        if aln.aligned:
            yield aln.iv

def gff_iv_iterator(gtf_file):
    for feature in HTSeq.GFF_Reader(gtf_file):
        yield feature.iv




import os
import unittest
from bin.pyreference_gff_to_json import parser_factory


class Test(unittest.TestCase):
    base_dir = os.path.join(os.path.dirname(__file__), "..")
    ENSEMBL_GTF_FILENAME = os.path.join(base_dir, "pyreference", "tests", "reference", "ensembl_test.GRCh38.104.gtf")
    REFSEQ_GFF3_FILENAME = os.path.join(base_dir, "pyreference", "tests", "reference", "refseq_test.GRCh38.p13_genomic.109.20210514.gff")
    UCSC_GTF_FILENAME = os.path.join(base_dir, "pyreference", "tests", "reference", "hg19_chrY_300kb_genes.gtf")

    def _test_exon_length(self, data, transcript_id, expected_length):
        transcript = data["transcripts_by_id"][transcript_id]
        exons = transcript["features_by_type"]["exon"]
        length = sum([d["stop"] - d["start"] for d in exons])
        self.assertEquals(expected_length, length, "%s exons sum" % transcript_id)


    def test_ucsc_gtf(self):
        parser = parser_factory(gtf=self.UCSC_GTF_FILENAME)
        data = parser.get_data()
        self._test_exon_length(data, "NM_013239", 2426)

    def test_ensembl_gtf(self):
        parser = parser_factory(gtf=self.ENSEMBL_GTF_FILENAME)
        data = parser.get_data()
        self._test_exon_length(data, "ENST00000357654.9", 7088)


    def test_refseq_gff3(self):
        parser = parser_factory(gff3=self.REFSEQ_GFF3_FILENAME)
        data = parser.get_data()
        self._test_exon_length(data, "NM_007294.4", 7088)
